#!/usr/bin/env python 
#
# (c) Peter Beerli
# MIT Licence
#
# minimal implementation of P. Green example 
# some arrays must be numpy arrays
#
import numpy as np
import math
import random as ra
import sys
from scipy.linalg import lu


def generate_data(elements,lamda,kappa):
    v = lamda + lamda*lamda/kappa
    n = -(lamda*lamda/(lamda - v))
    p = lamda/v
    return np.random.negative_binomial(n, p, elements)


def loggamma(x):
    return logfactorial(x-1.0)

def logfactorial(x):
    if isinstance(x, (list, tuple, np.ndarray)):
        return [math.lgamma(xi+1.0) for xi in x]
    else:
        return math.lgamma(x+1.0)

def logpoissonlike(Y,lamda):
    return np.sum(np.log(lamda) * Y  - logfactorial(Y) - lamda)

# kappa is 1/kappa in the Green and Hiena text 
def lognegativebinomiallike(Y,lamda,kappa):
    a = np.log(lamda) * Y  - logfactorial(Y)
    b = loggamma(kappa + Y) 
    c = loggamma(kappa) + np.log(kappa + lamda) * Y
    d = np.log(1.0 + lamda / kappa) * (-kappa)
    return np.sum(a + b - c + d) 

def logpdfgamma(x,a,b):
    #log ( b^a/Gamma(a) x^(a-1) exp(-b x) )
    # log(b)*a - loggama(a) + log(x)*(a-1) - b x
    return np.log(b)*a - loggamma(a) + np.log(x) * (a-1.0) - b*x

alphaL = 2.50
betaL =  1.0
def logp0(theta):
    return logpdfgamma(theta[0],alphaL,betaL)

alphaK = 1.0
betaK = 1.0
def logp1(theta):
    return logpdfgamma(theta[0],alphaL,betaL)+logpdfgamma(theta[1],alphaK,betaK)
    
loghalf = np.log(0.5)
def posterior(k, lamda, kappa, Y):
    if k==0:
        return loghalf + logp0([lamda]) + logpoissonlike(Y,lamda) 
    else:
        return loghalf + logp1([lamda,kappa]) + lognegativebinomiallike(Y,lamda,kappa) 

def proposal(x):
    global delta
    if isinstance(x, (list, tuple, np.ndarray)):
        x = np.fabs(x + (np.random.uniform(0,1)-0.5) * delta)
        if x[0]>100 or x[1]>100:
           x = np.remainder(x,100) 
    else:
        x = np.fabs(x + (np.random.uniform(0,1)-0.5) * delta)
        if x>100:
           x = np.remainder(x,100) 
    return x

def hastingsF(state1, state2, theta, mu, var, u):
    if state1==0:
        return -(np.log(np.sqrt(2*math.pi*var))-(u*u/(2*var)))
    else:
        xx = np.log(theta[1]/mu)
        return (-np.log(np.sqrt(2*math.pi*var))) - (xx*xx/(2*var))

def det(A):
    L,U = lu(A)
    return np.prod(np.diag(L)) * np.prod(np.diag(U))


def move(state,theta,Y):
    newtheta = np.array(theta[:])
    newtheta[1] = 1./newtheta[1]
    new = proposal(newtheta)
    new[1] = 1./ new[1]
    metro = posterior(state,new[0],new[1], Y) - posterior(state,theta[0],theta[1], Y)
    # log hastings is zero for this move
    #print "MOVE",new,theta
    rr = np.log(np.random.uniform(0,1))
    #rr = -100000000000
    if rr < metro:
        return True, new
    else:
        return False, theta


def jump(state,theta,Y, mu, var):
    if state == 0:
        r = 1
    else:
        r = 0
    #print "state=",state,"r=",r, 
    metro = posterior(r,theta[0],theta[1], Y) - posterior(state,theta[0],theta[1], Y)
    u = np.random.uniform(0.,1.) 
    if state == 0:
        hastings = hastingsF(0,1,theta,mu,var,u)
        green = np.log(mu) * u
    else:
        hastings = hastingsF(1,0, theta, mu,var,u)
        green = -np.log(theta[1]) 
    rr = np.log(np.random.uniform(0.,1.))
    #print "JUMP",r, state, metro, hastings, green, metro+hastings+green, rr < metro+hastings+green
    #rr = -100000000000
    if rr < metro+hastings+green:
        #print "YEAHH"
        return True, r
    else:
        #print " "
        return False, state

def runone(Y,theta,state,mu,var):
    global result
    #print "runone",state
    r = np.random.uniform(0,1)
    if r < 0.5:
        accept, newtheta = move(state,theta,Y)
        theta = newtheta[:]
        result.append([theta,state,accept,"M"])
    else:
        accept, newstate = jump(state,theta,Y,mu,var)
        state = newstate
        result.append([theta,state,accept,"J"])
    #print result[-1]
    return theta,state

def report_to_file(result,filename):
    f = open(filename,'w')
    for i in result:
        for ii in i:
            f.write(str(ii))
            f.write(' ')
        f.write('\n')
    f.close()

def report_rjmcmc(result):
    accm = [r[2] for r in result if r[-1]=='M']
    accj = [r[2] for r in result if r[-1]=='J']
    acceptmoves = len(filter(lambda x: x==True,accm))
    acceptjumps = len(filter(lambda x: x==True, accj))
    rejectmoves = len(filter(lambda x: x==False,accm))
    rejectjumps = len(filter(lambda x: x==False, accj))
    nn = acceptmoves + acceptjumps + rejectmoves + rejectjumps
    lamda1 = [r[0][0] for r in result if r[1]==0]
    lamda2 = [r[0][0] for r in result if r[1]==1]
    kappa = [1.0/r[0][1]  for r in result if r[1]==1]
    print "Model report:"
    try:
        len1 = len(lamda1)
        print "Poisson [freq]   :", len1/float(nn)
        print "           lambda:", np.mean(lamda1)  
    except:
        len1 = 0
        lamda1 = [0]
    try:
        len2 = len(lamda2)
        print "Negative Binomial:", len2/float(nn)
        print "           lambda:", np.mean(lamda2)  
    except:
        len2 = 0
        lamda2=[0]
    try:
        lenk = len(kappa)
        print "            kappa:", np.mean(kappa)  
        m = np.mean(lamda2)
        variance = m + m*m * np.mean(kappa)
        print "         Variance:", variance  
    except:
        lenk = 0
        kappa=[0]
    print
    print "Moves accepted:  %.3f N=%li" % (float(acceptmoves)/(acceptmoves+rejectmoves), acceptmoves+rejectmoves)
    print "Jumps accepted:  %.3f N=%li" % (float(acceptjumps)/(acceptjumps+rejectjumps), acceptjumps+rejectjumps)
    print
    print
    return len1,lamda1,len2,lamda2,lenk,kappa

def plot_rjmcmc(len1,lamda1,len2,lamda2,lenk,kappa):
    import matplotlib.pyplot as plt
    #print len(lamda1),len(lamda2), len(kappa)
    fig = plt.figure()
    ax1 = fig.add_subplot(2,2,1)
    ax2 = fig.add_subplot(2,2,3)
    ax3 = fig.add_subplot(2,2,4)

    fig.subplots_adjust(hspace=1)
    if len1 > 0:
        ax1.hist(lamda1,bins=25,edgecolor='None')
    if len2 > 0:
        ax2.hist(lamda2,bins=25,edgecolor='None')
    if lenk > 0:
        ax3.hist(kappa,bins=25,edgecolor='None')
    plt.show()


delta = 1.0
result = []
def main(args):
    global result
    global  delta
    lamda = 1.0
    kappa = 1/0.1
    nn = 10000
    # NB lambda=1,  (Variance=1.25) variance = lambda + lambda^2/kappa
    #    kappa= -(m^2/(m - v)) = 4 [?]
    Y = [0, 1, 0, 1, 4, 2, 1, 3, 3, 2, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 2, 2, 3, 2, 0, 1, 1, 2, 0, 1, 0, 1, 0, 1, 2, 0, 2, 0, 1, 1, 0, 0, 0, 2, 0, 0, 1, 1, 3, 1, 3, 0, 2, 0, 0, 2, 1, 2, 1, 0, 1, 2, 0, 0, 0, 0, 0, 2, 0, 1, 2, 0, 2, 1, 0, 0, 1, 0, 0, 2, 1, 0, 2, 0, 1, 3, 1, 2, 0, 2, 7, 1, 1, 0, 0]
    if len(sys.argv)>1:
        if 'h' in sys.argv[1]:
            print "Syntax: python rj.py #sweeps < dataelements datamean datavariance"
            print "        Poisson: datamean=datavariance"
            print "        NegativeBinomial: datamean and datavariance are independent"
            sys.exit()
        else:
            nn = int(args[1])
    if len(args)>5:
        elements = int(args[2])
        lamda = float(args[3])
        kappa = float(args[4])
        Y=generate_data(lamda,kappa,elements)
    Y = np.array(Y)
    mean = np.mean(Y)
    var = np.var(Y)
    mu = 1/(np.sqrt(2*var*np.pi))
    delta = mu
    print "From data:" 
    print "Mean     :",mean
    print "Variance :",var
    print "Tuning parameters:"
    print "mu               :",mu
    print "delta            :",delta
    print
    theta = [lamda,kappa]
    state = 0
    result = []
    for i in xrange(nn):
        theta,state = runone(Y,theta,state,mu,var)
    report_to_file(result,"temp")
    len1, lamda1, len2, lamda2, lenk, kappa = report_rjmcmc(result)
    plot_rjmcmc(len1,lamda1,len2,lamda2,lenk,kappa)

if __name__ == '__main__':
    main(sys.argv)
 
    
